package optional

import (
	"encoding/json"
	"fmt"

	"github.com/marcsantiago/gojay"
)

type (
	optionalConstraint interface {
		~int | ~int8 | ~int16 | ~int32 | ~int64 | ~float32 | ~float64 | ~string
	}
)

// Optional wraps value types with a set flag
type Optional[T optionalConstraint] struct {
	value T
	isSet bool
}

// New instantiates a new optional with a concrete
func New[T optionalConstraint](value T) Optional[T] {
	return Optional[T]{
		value: value,
		isSet: true,
	}
}

// Set ensures that the internal value is set explicitly
func (o *Optional[T]) Set(value T) {
	o.value = value
	o.isSet = true
}

// Get returns the concrete value without checking if the value has been set
func (o *Optional[T]) Get() (value T) {
	return o.value
}

// IsSome returns true if the value have been set either by calling New or Set
func (o *Optional[T]) IsSome() bool {
	return o.isSet
}

// IsNone returns true if the internal value has not been explicitly Set
func (o *Optional[T]) IsNone() bool {
	return !o.isSet
}

// Unwrap returns the underline concrete value and a boolean indicating if the value has been set or not
func (o *Optional[T]) Unwrap() (value T, hasBeenSet bool) {
	return o.value, o.isSet
}

// GoJay serialization methods

func (o *Optional[T]) MarshalJSONObject(enc *gojay.Encoder) {
	if o.isSet {
		// using a tiny bit of reflection to avoid calling enc.Encode(v interface{}), which will cause value to escape to the
		// heap TODO I can extend GoJay to accept a type T which will compile out the static calls for Nimbus
		// for now the constraint is small enough that this is fine, the cases are also exhaustive, given T is constrained already
		switch vt := any(o.value).(type) {
		case int:
			enc.Int(vt)
		case int8:
			enc.Int8(vt)
		case int16:
			enc.Int16(vt)
		case int32:
			enc.Int32(vt)
		case int64:
			enc.Int64(vt)
		case float32:
			enc.Float32(vt)
		case float64:
			enc.Float64(vt)
		case string:
			enc.String(vt)
		}
	}
}

func (o *Optional[T]) IsNil() bool {
	return !o.isSet
}

func (o *Optional[T]) UnmarshalJSONObject(dec *gojay.Decoder, _ string) error {
	err := dec.Decode(&o.value)
	if err == nil {
		o.isSet = true
	}
	return err
}

func (o *Optional[T]) NKeys() int { return 0 }

// standard library methods
// Known limitation in the Go standard library, where one can't simply omit a value from MarshalJSON
// https://github.com/golang/go/issues/50480
// https://github.com/golang/go/issues/45669

// Note the per Go's standard implementations all the methods case V below escape to the heap because of the untyped interface{} value

func (o *Optional[T]) MarshalJSON() ([]byte, error) {
	if o.isSet {
		return json.Marshal(o.value)
	}
	// null has to be used to avoid Go internal limitation outlined in the above issues
	return []byte("null"), nil
}

func (o *Optional[T]) UnmarshalJSON(data []byte) error {
	err := json.Unmarshal(data, &o.value)
	if err == nil {
		o.isSet = true
	}
	return err
}

// String to satisfy the Stringer interface for printing
func (o *Optional[T]) String() string {
	return fmt.Sprintf("%v", o.value)
}
