package optional

import (
	"bytes"
	"strings"
	"testing"

	testutil "github.com/adsbynimbus/nimbus/internal/test-util"
	"github.com/google/go-cmp/cmp"
	"github.com/marcsantiago/gojay"
)

func TestOptional_Get(t *testing.T) {
	type args struct {
		value float64
	}
	type testCase[T optionalConstraint] struct {
		name      string
		args      args
		wantValue T
	}
	tests := []testCase[float64]{
		{
			name:      "should return 5.0",
			wantValue: 5.0,
			args: args{
				value: 5.0,
			},
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			opt := New(tt.args.value)
			if gotValue := opt.Get(); gotValue != tt.wantValue {
				t.Errorf("Get() = %v, want %v", gotValue, tt.wantValue)
			}
		})
	}
}

func TestOptional_None(t *testing.T) {
	type testCase[T optionalConstraint] struct {
		name string
		o    Optional[T]
		want bool
	}
	tests := []testCase[int]{
		{
			name: "should return true, no value as been set",
			o:    Optional[int]{},
			want: true,
		},
		{
			name: "should return false, no an integer has been set",
			o: Optional[int]{
				value: 55,
				isSet: true,
			},
			want: false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if got := tt.o.IsNone(); got != tt.want {
				t.Errorf("IsNone() = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestOptional_Some(t *testing.T) {
	type testCase[T optionalConstraint] struct {
		name string
		o    Optional[T]
		want bool
	}
	tests := []testCase[string]{
		{
			name: "should return false, no value as been set",
			o:    Optional[string]{},
			want: false,
		},
		{
			name: "should return true, no an integer has been set",
			o: Optional[string]{
				value: "foobar",
				isSet: true,
			},
			want: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if got := tt.o.IsSome(); got != tt.want {
				t.Errorf("IsSome() = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestOptional_Unwrap(t *testing.T) {
	type testCase[T optionalConstraint] struct {
		name      string
		o         Optional[T]
		wantValue T
		wantOk    bool
	}
	tests := []testCase[int64]{
		{
			name: "should return the value with the flag set",
			o: Optional[int64]{
				value: 30,
				isSet: true,
			},
			wantValue: 30,
			wantOk:    true,
		},
		{
			name:      "should return false and the primitive's default value",
			o:         Optional[int64]{},
			wantValue: 0,
			wantOk:    false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			gotValue, gotOk := tt.o.Unwrap()
			if gotValue != tt.wantValue {
				t.Errorf("Unwrap() gotValue = %v, want %v", gotValue, tt.wantValue)
			}
			if gotOk != tt.wantOk {
				t.Errorf("Unwrap() gotOk = %v, want %v", gotOk, tt.wantOk)
			}
		})
	}
}

type testData struct {
	Marker   string
	BidValue Optional[float64]
}

func (t *testData) MarshalJSONObject(enc *gojay.Encoder) {
	enc.StringKey("marker", t.Marker)
	if value, ok := t.BidValue.Unwrap(); ok {
		enc.Float64KeyOmitEmpty("bid_value", value)
	}
}

func (t *testData) IsNil() bool {
	return len(t.Marker) == 0 && t.BidValue.IsNone()
}

func (t *testData) UnmarshalJSONObject(dec *gojay.Decoder, k string) error {
	switch k {
	case "marker":
		_ = dec.String(&t.Marker)

	case "bid_value":
		var bid float64
		_ = dec.Float64(&bid)
		t.BidValue.Set(bid)
	}
	return nil
}

func (t *testData) NKeys() int { return 0 }

func TestOptional_Encode(t *testing.T) {
	type testCase struct {
		name     string
		testData testData
		want     string
		wantErr  bool
	}
	tests := []testCase{
		{
			name: "serialize the optional data",
			testData: testData{
				Marker: "marker",
				BidValue: Optional[float64]{
					value: 1.87948,
					isSet: true,
				},
			},
			want:    "{\"marker\":\"marker\",\"bid_value\":1.87948}",
			wantErr: false,
		},
		{
			name: "serialize the just the marker data, isSet is set to false",
			testData: testData{
				Marker: "marker",
				BidValue: Optional[float64]{
					value: 25.04,
					isSet: false,
				},
			},
			want:    "{\"marker\":\"marker\"}",
			wantErr: false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			buf := new(bytes.Buffer)
			enc := gojay.NewEncoder(buf)
			gotErr := enc.EncodeObject(&tt.testData)
			if tt.wantErr && gotErr == nil {
				t.Fatalf("Encode() wanted an error, got none")
			}
			if tt.want != buf.String() {
				t.Fatalf("Encode() want %s got %s", tt.want, buf.String())
			}
		})
	}
}

func TestOptional_Decode(t *testing.T) {
	type testCase struct {
		name    string
		input   string
		want    testData
		wantErr bool
	}
	tests := []testCase{
		{
			name:  "should deserialize the optional data",
			input: "{\"marker\":\"marker\",\"bid_value\":1.87948}",
			want: testData{
				Marker: "marker",
				BidValue: Optional[float64]{
					value: 1.87948,
					isSet: true,
				},
			},
			wantErr: false,
		},
		{
			name:  "should not deserialize the optional data",
			input: "{\"marker\":\"marker\"}",
			want: testData{
				Marker:   "marker",
				BidValue: Optional[float64]{},
			},
			wantErr: false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			dec := gojay.NewDecoder(strings.NewReader(tt.input))
			var got testData
			gotErr := dec.DecodeObject(&got)
			if tt.wantErr && gotErr == nil {
				t.Fatalf("Decode() wanted an error, got none")
			}
			if !cmp.Equal(tt.want, got, testutil.DeepAllowUnexported(tt.want, got)) {
				t.Fatalf("Decode() want %+v got %+v", tt.want, got)
			}
		})
	}
}
